# Copyright (C) 2010 W. Trevor King <wking@drexel.edu>
#
# This file is part of Hooke.
#
# Hooke is free software: you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation, either
# version 3 of the License, or (at your option) any later version.
#
# Hooke is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Hooke.  If not, see
# <http://www.gnu.org/licenses/>.

"""The `playlist` module provides a :class:`Playlist` and its subclass
:class:`FilePlaylist` for manipulating lists of
:class:`hooke.curve.Curve`\s.
"""

import copy
import hashlib
import os.path
import xml.dom.minidom

from . import curve as curve


class NoteIndexList (list):
    """A list that keeps track of a "current" item and additional notes.

    :attr:`index` (i.e. "bookmark") is the index of the currently
    current curve.  Also keep a :class:`dict` of additional information
    (:attr:`info`).
    """
    def __init__(self, name=None):
        super(NoteIndexList, self).__init__()
        self.name = name
        self.info = {}
        self._index = 0

    def __str__(self):
        return '<%s %s>' % (self.__class__.__name__, self.name)

    def _setup_item(self, item):
        """Perform any required initialization before returning an item.
        """
        pass

    def current(self):
        if len(self) == 0:
            return None
        item = self[self._index]
        self._setup_item(item)
        return item

    def jump(self, index):
        if len(self) == 0:
            self._index = 0
        else:
            self._index = index % len(self)

    def next(self):
        self.jump(self._index + 1)

    def previous(self):
        self.jump(self._index - 1)

    def filter(self, keeper_fn=lambda item:True, *args, **kwargs):
        c = copy.deepcopy(self)
        for item in reversed(c):
            c._setup_item(item)
            if keeper_fn(item, *args, **kwargs) != True:
                c.remove(item)
        try: # attempt to maintain the same current item
            c._index = c.index(self.current())
        except ValueError:
            c._index = 0
        return c

class Playlist (NoteIndexList):
    """A :class:`NoteIndexList` of :class:`hooke.curve.Curve`\s.

    Keeps a list of :attr:`drivers` for loading curves.
    """
    def __init__(self, drivers, name=None):
        super(Playlist, self).__init__(name=name)
        self.drivers = drivers
        self._loaded = [] # List of loaded curves, see :meth:`._setup_item`.
        self._max_loaded = 100 # curves to hold in memory simultaneously.

    def append_curve_by_path(self, path, info=None, identify=True):
        if self.path != None:
            path = os.path.join(os.path.dirname(self.path), path)
        path = os.path.normpath(path)
        c = curve.Curve(path, info=info)
        if identify == True:
            c.identify(self.drivers)
        self.append(c)
        return c

    def _setup_item(self, curve):
        if curve != None and curve not in self._loaded:
            if curve not in self:
                self.append(curve)
            if curve.driver == None:
                c.identify(self.drivers)
            if curve.data == None:
                curve.load()
            self._loaded.append(curve)
            if len(self._loaded) > self._max_loaded:
                oldest = self._loaded.pop(0)
                oldest.unload()

class FilePlaylist (Playlist):
    version = '0.1'

    def __init__(self, drivers, name=None, path=None):
        super(FilePlaylist, self).__init__(drivers, name)
        self.set_path(path)
        self._digest = None

    def set_path(self, path):
        if path != None:
            if not path.endswith('.hkp'):
                path += '.hkp'
            self.path = path
            if self.name == None:
                self.name = os.path.basename(path)

    def is_saved(self):
        return self.digest() == self._digest

    def digest(self):
        r"""Compute the sha1 digest of the flattened playlist
        representation.

        Examples
        --------

        >>> root_path = os.path.sep + 'path'
        >>> p = FilePlaylist(drivers=[],
        ...                  path=os.path.join(root_path, 'to','playlist'))
        >>> p.info['note'] = 'An example playlist'
        >>> c = curve.Curve(os.path.join(root_path, 'to', 'curve', 'one'))
        >>> c.info['note'] = 'The first curve'
        >>> p.append(c)
        >>> c = curve.Curve(os.path.join(root_path, 'to', 'curve', 'two'))
        >>> c.info['note'] = 'The second curve'
        >>> p.append(c)
        >>> p.digest()
        '\\\x14\x87\x88*q\xf8\xaa\xa7\x84f\x82\xa1S>\xfd3+\xd0o'
        """
        string = self.flatten()
        return hashlib.sha1(string).digest()

    def flatten(self, absolute_paths=False):
        """Create a string representation of the playlist.

        A playlist is an XML document with the following syntax::

            <?xml version="1.0" encoding="utf-8"?>
            <playlist attribute="value">
              <curve path="/my/file/path/"/ attribute="value" ...>
              <curve path="...">
            </playlist>

        Relative paths are interpreted relative to the location of the
        playlist file.
        
        Examples
        --------

        >>> root_path = os.path.sep + 'path'
        >>> p = FilePlaylist(drivers=[],
        ...                  path=os.path.join(root_path, 'to','playlist'))
        >>> p.info['note'] = 'An example playlist'
        >>> c = curve.Curve(os.path.join(root_path, 'to', 'curve', 'one'))
        >>> c.info['note'] = 'The first curve'
        >>> p.append(c)
        >>> c = curve.Curve(os.path.join(root_path, 'to', 'curve', 'two'))
        >>> c.info['note'] = 'The second curve'
        >>> p.append(c)
        >>> print p.flatten() # doctest: +NORMALIZE_WHITESPACE +REPORT_UDIFF
        <?xml version="1.0" encoding="utf-8"?>
        <playlist index="0" note="An example playlist" version="0.1">
            <curve note="The first curve" path="curve/one"/>
            <curve note="The second curve" path="curve/two"/>
        </playlist>
        <BLANKLINE>
        >>> print p.flatten(absolute_paths=True) # doctest: +NORMALIZE_WHITESPACE +REPORT_UDIFF
        <?xml version="1.0" encoding="utf-8"?>
        <playlist index="0" note="An example playlist" version="0.1">
            <curve note="The first curve" path="/path/to/curve/one"/>
            <curve note="The second curve" path="/path/to/curve/two"/>
        </playlist>
        <BLANKLINE>
        """
        implementation = xml.dom.minidom.getDOMImplementation()
        # create the document DOM object and the root element
        doc = implementation.createDocument(None, 'playlist', None)
        root = doc.documentElement
        root.setAttribute('version', self.version) # store playlist version
        root.setAttribute('index', str(self._index))
        for key,value in self.info.items(): # save info variables
            root.setAttribute(key, str(value))
        for curve in self: # save curves and their attributes
            curve_element = doc.createElement('curve')
            root.appendChild(curve_element)
            path = os.path.abspath(os.path.expanduser(curve.path))
            if absolute_paths == False:
                path = os.path.relpath(
                    path,
                    os.path.dirname(
                        os.path.abspath(
                            os.path.expanduser(self.path))))
            curve_element.setAttribute('path', path)
            for key,value in curve.info.items():
                curve_element.setAttribute(key, str(value))
        string = doc.toprettyxml(encoding='utf-8')
        root.unlink() # break circular references for garbage collection
        return string

    def _from_xml_doc(self, doc, identify=True):
        """Load a playlist from an :class:`xml.dom.minidom.Document`
        instance.
        """
        root = doc.documentElement
        for attribute,value in root.attributes.items():
            if attribute == 'version':
                assert value == self.version, \
                    'Cannot read v%s playlist with a v%s reader' \
                    % (value, self.version)
            elif attribute == 'index':
                self._index = int(value)
            else:
                self.info[attribute] = value
        for curve_element in doc.getElementsByTagName('curve'):
            path = curve_element.getAttribute('path')
            info = dict(curve_element.attributes.items())
            info.pop('path')
            self.append_curve_by_path(path, info, identify=identify)
        self.jump(self._index) # ensure valid index

    def from_string(self, string, identify=True):
        """Load a playlist from a string.

        Examples
        --------

        >>> string = '''<?xml version="1.0" encoding="utf-8"?>
        ... <playlist index="1" note="An example playlist" version="0.1">
        ...     <curve note="The first curve" path="../curve/one"/>
        ...     <curve note="The second curve" path="../curve/two"/>
        ... </playlist>
        ... '''
        >>> p = FilePlaylist(drivers=[],
        ...                  path=os.path.join('path', 'to', 'my', 'playlist'))
        >>> p.from_string(string, identify=False)
        >>> p._index
        1
        >>> p.info
        {u'note': u'An example playlist'}
        >>> for curve in p:
        ...     print curve.path
        path/to/curve/one
        path/to/curve/two
        """
        doc = xml.dom.minidom.parseString(string)
        self._from_xml_doc(doc, identify=identify)

    def load(self, path=None, identify=True):
        """Load a playlist from a file.
        """
        self.set_path(path)
        doc = xml.dom.minidom.parse(self.path)
        self._from_xml_doc(doc, identify=identify)
        self._digest = self.digest()

    def save(self, path=None):
        """Saves the playlist in a XML file.
        """
        self.set_path(path)
        f = file(self.path, 'w')
        f.write(self.flatten())
        f.close()
